from flexible_permissions._utils import ensure_plural, is_value

"""
Maps of roles to actions
"""

ROLES = {}
ACTIONS = {}


def calculate_actions(roles):
    result = {}
    for role, actions in roles.items():
        for action in actions:
            result.setdefault(action, [])
            result[action].append(role)
            result[action] = list(set(result[action]))

    return result


def register_role(name, actions):
    global ACTIONS

    for action in actions:
        if action.count(".") != 1:
            raise ValueError(
                "Actions should have one period in them - the left portion"
                "should be the name of the object; the right the name of the"
                "action."
            )

    # Merge the actions together
    pre_existing_actions = ROLES.get(name, [])
    ROLES[name] = list(set(actions + pre_existing_actions))

    # Keep action map in sync
    ACTIONS = calculate_actions(ROLES)


def roles_to_actions(roles):
    # Keep special values
    if not is_value(roles):
        return roles

    roles = ensure_plural(roles)
    return list(set([action for role in roles for action in ROLES[role]]))


def actions_to_roles(actions):
    # Keep special values
    if not is_value(actions):
        return actions

    actions = ensure_plural(actions)
    return list(set([role for action in actions for role in ACTIONS[action]]))



# -*- coding: utf-8 -*-
from __future__ import unicode_literals

from django.db import models, migrations


class Migration(migrations.Migration):

    dependencies = [
        ('contenttypes', '0001_initial'),
    ]

    operations = [
        migrations.CreateModel(
            name='Permission',
            fields=[
                ('id', models.AutoField(verbose_name='ID', serialize=False, auto_created=True, primary_key=True)),
                ('role', models.CharField(max_length=255)),
                ('agent_id', models.PositiveIntegerField(null=True, blank=True)),
                ('target_id', models.PositiveIntegerField()),
                ('agent_type', models.ForeignKey(related_name='+', on_delete=models.deletion.PROTECT, blank=True, to='contenttypes.ContentType', null=True)),
                ('target_type', models.ForeignKey(related_name='+', on_delete=models.deletion.PROTECT, to='contenttypes.ContentType')),
            ],
        ),
        migrations.AlterUniqueTogether(
            name='permission',
            unique_together=set([('role', 'agent_type', 'agent_id', 'target_type', 'target_id')]),
        ),
    ]



from django.db import models
from django.contrib.contenttypes.fields import GenericForeignKey
from django.contrib.contenttypes.models import ContentType


class Permission(models.Model):
    role = models.CharField(max_length=255)

    agent_type = models.ForeignKey(
        ContentType,
        related_name='+',
        on_delete=models.PROTECT,
        null=True,
        blank=True
    )
    agent_id = models.PositiveIntegerField(null=True, blank=True)
    agent = GenericForeignKey('agent_type', 'agent_id')

    target_type = models.ForeignKey(
        ContentType,
        related_name='+',
        on_delete=models.PROTECT
    )
    target_id = models.PositiveIntegerField()
    target = GenericForeignKey('target_type', 'target_id')

    class Meta:
        unique_together = [
            'role',
            'agent_type',
            'agent_id',
            'target_type',
            'target_id'
        ]

    def __unicode__(self):
        return "%s %s" % (self.__class__.__name__, self.pk)



from django.db.models import Q, QuerySet

from itertools import chain

from flexible_permissions.agents import normalize_agent
from flexible_permissions.relations import (
    get_related_target_prefixes,
    get_related_agent_prefixes,
)
from flexible_permissions.roles import actions_to_roles
from flexible_permissions._utils import (
    ANY,
    NULL,
    ensure_plural,
    generic_in,
    get_key,
    filter_isnull,
    is_value,
    normalize_value,
    get_model_name,
    validate_roles_with_targets,
)


def define_filter(key):
    def decorator(fn):
        def wrapped(queryset, value=ANY, prefix=None):
            if value is ANY:
                return Q()

            if value is NULL:
                return filter_isnull(key, prefix)

            # The fn should return a Q object.
            return fn(queryset, **{key: value, 'prefix': prefix})

        return wrapped

    return decorator


class PermQuerySet(QuerySet):

    """
    Query Building stuff
    """

    def _get_filter(self, key, value):
        return generic_in(key, ensure_plural(value))

    @define_filter('role')
    def _get_role_query(self, role, prefix):
        return Q(**{get_key('role', prefix, 'in'): role})

    @define_filter('agent')
    def _get_agent_query(self, agent, prefix):
        # Always add in isnull
        return (
            filter_isnull('agent', prefix) |
            self._get_filter(get_key('agent', prefix), agent)
        )

    @define_filter('target')
    def _get_target_query(self, target, prefix):
        return self._get_filter(get_key('target', prefix), target)

    def _get_query(
        self,
        role=ANY,
        agent=ANY,
        target=ANY,
        prefix=None
    ):
        """
        This constructs a query for the filter on permissions.
        """
        return (
            self._get_role_query(role, prefix) &
            self._get_agent_query(agent, prefix) &
            self._get_target_query(target, prefix)
        )

    """
    Query execution
    """

    def _query_separate(self, queries):
        """
        Combining too many queries can cause horrendous LEFT JOINS
        and we end up scanning millions of rows. Instead, query each
        separately as ids, then query again to get all objects
        """

        # Reset ordering to avoid extra joins
        queryset = self.order_by('pk')

        ids = []
        for query in queries:
            ids.append(queryset.filter(query).values_list('id', flat=True))

        # Keep the ordering here since this is the result
        return self.filter(id__in=set(chain(*ids)))

    def _query_together(self, queries):
        """
        This queries everything all at once. This is safe if we're joining
        tables in only a couple different directions.
        """
        return reduce(
            lambda queryset, query: queryset | self.filter(query),
            queries,
            self.none()
        )

    def _query_perms(self,
        roles,
        get_related_prefixes,
        perms_name,
        force_separate,
        agent=ANY,
        target=ANY
    ):
        """
        This is an abstraction used by subclasses to query permissions.

        Roles will be normalized to a list of roles.

        force_separate is an optimization. If you know you are going
        to have multiple left joins, this will speed things up.

        Either agent or target can be provided. It's assumed that the
        queryset to be retrieved is the thing not provided.
        """
        # Normalize inputs
        roles = normalize_value(roles)
        agent = normalize_value(agent)
        target = normalize_value(target)

        # Get all possible related queries we're doing
        related_prefixes = (
            get_related_prefixes(self, perms_name, *roles)
            if is_value(roles) else
            [perms_name]
        )

        # Create a query for each related prefix
        queries = [
            self._get_query(roles, agent, target, prefix=prefix)
            for prefix in related_prefixes
        ]

        # Aggregate the queries. Query together if we don't have any
        # divergent left joins.
        query_together = (
            not force_separate and
            len(related_prefixes) <= 1 and
            perms_name in related_prefixes
        )

        if query_together:
            results = self._query_together(queries)
        else:
            results = self._query_separate(queries)

        return results


class PermTargetQuerySet(PermQuerySet):
    """
    This is a queryset class that exposes methods to get a related target
    via the permissions table. The methods should be read as
    "Get a target for an agent with role x."
    """
    def _get_model_name(self):
        """
        Make this overridable
        """
        return get_model_name(self.model)

    def _prefix_actions(self, actions):
        """
        Prepend this model's name as prefix. This enforces more correct
        configuration when registering role/action mappings.
        """
        if not is_value(actions):
            return actions

        result = []
        for action in ensure_plural(actions):
            if "." in action:
                raise ValueError(
                    "Prefixes are inferred. Register this with its own "
                    "actions."
                )

            result.append("%s.%s" % (self._get_model_name(), action))

        return result

    def for_role(
        self,
        roles=ANY,
        agent=ANY,
        infer_agents=True,
        force_separate=False
    ):
        """
        This filters permission targets by the given agent.

        infer_agents is an optimization. If you know you don't need the
        authority of any related agents, set it to false.
        """
        return self._query_perms(
            roles=roles,
            get_related_prefixes=get_related_target_prefixes,
            perms_name='target_perms',
            force_separate=force_separate,
            agent=normalize_value(
                agent,
                normalize_agent,
                infer_agents=infer_agents
            )
        )

    def for_action(self, actions=ANY, *args, **kwargs):
        roles = actions_to_roles(self._prefix_actions(actions))
        return self.for_role(roles, *args, **kwargs)


class PermAgentQuerySet(PermQuerySet):
    """
    This is a queryset class that exposes methods to get a related agent
    via the permissions table. The methods should be read as
    "Get an agent with role x and target y."
    """
    def _validate_actions(self, actions):
        if not is_value(actions):
            return actions

        actions = normalize_value(actions)

        if [action for action in actions if "." not in action]:
            raise ValueError("Prefixes are required since target is optional.")

        return actions

    def with_role(self, roles=ANY, target=ANY, force_separate=False):
        """
        This filters permission agents by the given target.
        """
        validate_roles_with_targets(roles, target)

        return self._query_perms(
            roles=roles,
            get_related_prefixes=get_related_agent_prefixes,
            perms_name='agent_perms',
            force_separate=force_separate,
            target=target
        )

    def with_action(self, actions=ANY, *args, **kwargs):
        roles = actions_to_roles(self._validate_actions(actions))
        return self.with_role(roles, *args, **kwargs)



from flexible_permissions._utils import ensure_plural

_agent_registry = {}


def register_agent(cls, get_related_agents):
    """
    Register class function pair to get related agents.
    get_related_agents takes an agent and returns other agents
    that that agent implies. This should be used for getting
    groups from a user, for example.
    """
    _agent_registry[cls] = get_related_agents


def normalize_agent(agents, infer_agents=True):
    """
    Turn an agent into a list of equivalent agents
    """
    result = []
    for agent in ensure_plural(agents):
        if agent.__class__ not in _agent_registry:
            raise KeyError("%s is not registered." % agent.__class__)

        # Delegate agent calculation to registered function
        if infer_agents:
            result += list(_agent_registry[agent.__class__](agent))
        else:
            result.append(agent)

    return list(set(result))



from django.contrib.contenttypes.models import ContentType
from django.db import models
from django.db.models import Q

import operator

"""
Unique values for differentiating between
not filtering and constraining to isnull
"""


class ANY(object):
    pass


class NULL(object):
    pass


"""
Basic utility functions
"""


def is_plural(value):
    """
    Whether a value is an iterable, excluding dicts.
    """
    return hasattr(value, '__iter__') and not isinstance(value, dict)


def identity(value):
    """
    Just yer regular old identity function
    """
    return value


def ensure_plural(value, remove_nones=True):
    """
    If value is singular, wraps it in an array.
    If None is in value and remove_nones is True, None is filtered out.
    """
    if not is_plural(value):
        value = [value]

    if remove_nones:
        value = filter(identity, value)

    return value


def get_model_class(value):
    if isinstance(value, ContentType):
        model_class = value.model_class()
    elif isinstance(value, models.Model):
        model_class = value.__class__
    elif isinstance(value, models.QuerySet):
        model_class = value.model
    elif issubclass(value, models.Model):
        model_class = value
    elif issubclass(value, models.BaseManager):
        model_class = value.model
    else:
        raise TypeError("Invalid value passed: %s" % value)

    return model_class


def get_model_name(value):
    return get_model_class(value)._meta.verbose_name.replace(' ', '_')

"""
Generic query building functions
"""


def get_key(key, prefix=None, suffix=None):
    prefix = "%s__" % prefix if prefix is not None else ''
    suffix = "__%s" % suffix if suffix is not None else ''

    return "%s%s%s" % (prefix, key, suffix)


def filter_isnull(key, prefix=None):
    # Gotta add id to avoid reverse relation errors
    if key != 'role':
        key += '_id'

    # Return just a null check
    return Q(**{get_key(key, prefix, 'isnull'): True})


def is_value(value):
    return value not in [NULL, ANY]


def normalize_value(value, fn=ensure_plural, *args, **kwargs):
    return fn(value, *args, **kwargs) if is_value(value) else value


def generic_in(key, items):
    """
    Creates a the equivalent of an __in query for the given items.
    Items may be heterogeneous, but they must be Django models.
    """
    clauses = [
        Q(**{
            "%s_id" % key: item.id,
            "%s_type" % key: ContentType.objects.get_for_model(item),
        })
        for item in items
    ]

    # Match nothing for base query
    return reduce(operator.or_, clauses, Q(**{'id__isnull': True}))


def validate_roles_with_targets(roles, targets):
    if not is_value(roles) or not is_value(targets):
        return

    for target in normalize_value(targets):
        target_name = get_model_name(target)

        for role in normalize_value(roles):
            role_name = role.split(".")[0]
            if role_name != target_name:
                raise ValueError("Role %s is invalid for %s" % (
                    role_name,
                    target_name
                ))


def get_single_crud_kwargs(role, agent, target):
    """
    Gets a dict for filtering Permissions by Role, Agent, and Target.
    All arguments are optional, but must be explicitly provided.
    """
    kwargs = {}

    if role is NULL:
        kwargs['role__isnull'] = True
    elif is_value(role):
        kwargs['role'] = role

    if agent is NULL:
        kwargs['agent_id__isnull'] = True
    elif is_value(agent):
        kwargs.update({
            'agent_type': ContentType.objects.get_for_model(agent),
            'agent_id': agent.id
        })

    if target is NULL:
        kwargs['target_id__isnull'] = True
    elif is_value(target):
        kwargs.update({
            'target_type': ContentType.objects.get_for_model(target),
            'target_id': target.id
        })

    return kwargs


def get_multi_crud_query(role=NULL, agent=NULL, target=NULL):
    """
    Gets a query object for the given role, agent, and target. Any of the
    three may be plural; a generic_in will be used.
    """
    query = Q()

    if role is NULL:
        query = query & Q(role__isnull=True)
    elif is_value(role):
        query = query & Q(role__in=ensure_plural(role))

    if agent is NULL:
        query = query & Q(agent_id__isnull=True)
    elif is_value(agent):
        query = query & generic_in('agent', ensure_plural(agent))

    if target is NULL:
        query = query & Q(target_id__isnull=True)
    elif is_value(target):
        query = query & generic_in('target', ensure_plural(target))

    return query



# Include ANY and NULL as part of the public api
from flexible_permissions._utils import (
    ANY,
    NULL,
    validate_roles_with_targets,
    get_multi_crud_query,
    get_single_crud_kwargs,
)
from flexible_permissions.models import Permission

"""
Get
"""


def get_perms(*args, **kwargs):
    """
    Gets all Permissions matching the query.
    Accepts role, agent, and target kwargs.
    """
    return Permission.objects.filter(get_multi_crud_query(*args, **kwargs))


"""
Create
"""


def add_perm(role, agent, target):
    """
    Adds a single Permission matching the arguments.
    Accepts role, agent, and target kwargs.
    If the Permission already exists, it will retrieve it instead.
    """
    # Make sure we're not adding bad data
    validate_roles_with_targets(role, target)

    # If it already exists, this is a duplicate, so ignore it.
    query_kwargs = get_single_crud_kwargs(role, agent, target)
    return Permission.objects.get_or_create(**query_kwargs)[0]


"""
Delete
"""


def remove_perm(*args, **kwargs):
    """
    Removes all Permissions matching the arguments.
    Accepts role, agent, and target kwargs.
    """
    query_kwargs = get_single_crud_kwargs(*args, **kwargs)
    Permission.objects.filter(**query_kwargs).delete()


"""
Update
"""


def assign_role(role, agent, target):
    """
    Removes role for target from all other agents and assigns it to the
    given agent.
    """
    remove_perm(role, ANY, target)
    add_perm(role, agent, target)



from django.contrib.contenttypes.fields import GenericRelation
from django.db.models import Model

from flexible_permissions.models import Permission
from flexible_permissions.query import (
    PermTargetQuerySet,
    PermAgentQuerySet,
)


class PermTarget(Model):
    target_perms = GenericRelation(
        Permission,
        object_id_field='target_id',
        content_type_field='target_type'
    )

    objects = PermTargetQuerySet.as_manager()

    class Meta:
        abstract = True


class PermAgent(Model):
    agent_perms = GenericRelation(
        Permission,
        object_id_field='agent_id',
        content_type_field='agent_type'
    )

    objects = PermAgentQuerySet.as_manager()

    class Meta:
        abstract = True



_relation_registry = {}


def register_relation(cls, paths):
    _relation_registry[cls] = paths


def get_relation_paths(cls):
    """
    Return paths registered for the given class or any of its subclasses
    """

    results = {}
    [
        results.update(paths)
        for registered_cls, paths in _relation_registry.iteritems()
        if cls == registered_cls or issubclass(cls, registered_cls)
    ]

    return results


def get_related_target_prefixes(queryset, perms_name, *roles):
    paths = get_relation_paths(queryset.model)
    prefixes = []

    if perms_name in queryset.model._meta.get_all_field_names():
        prefixes.append(perms_name)

    for role in roles:
        role_prefix = role.split(".")[0]

        # Add in the path to the related object, plus perm
        if role_prefix in paths:
            if not isinstance(paths[role_prefix], list):
                paths[role_prefix] = [paths[role_prefix]]
            for path in paths[role_prefix]:
                prefixes.append(path + '__' + perms_name)

    return set(prefixes)


def get_related_agent_prefixes(queryset, perms_name, *roles):
    prefixes = []

    if perms_name in queryset.model._meta.get_all_field_names():
        prefixes.append(perms_name)

    return set(prefixes)



import os
from setuptools import setup

with open(os.path.join(os.path.dirname(__file__), 'README.md')) as readme:
    README = readme.read()

# allow setup.py to be run from any path
os.chdir(os.path.normpath(os.path.join(os.path.abspath(__file__), os.pardir)))

setup(
    name='django-flexible-permissions',
    version='1.0.5',
    packages=[
        'flexible_permissions',
        'flexible_permissions.migrations',
    ],
    include_package_data=True,
    license='MIT License',
    description=(
        "Django app that combines a permissions table with model "
        "relations to calculate complex object-level permissions."
    ),
    long_description=README,
    url='https://github.com/staab/django-flexible-permissions',
    author='Jon Staab',
    author_email='shtaab@gmail.com',
    classifiers=[
        'Environment :: Web Environment',
        'Framework :: Django',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: MIT License',
        'Operating System :: OS Independent',
        'Programming Language :: Python',
        'Topic :: Internet :: WWW/HTTP',
        'Topic :: Internet :: WWW/HTTP :: Dynamic Content',
    ],
)
